#!/usr/bin/env python3

__license__ = """
Simple Parser
Copyright (C) 2009 Christophe Delord
http://christophe.delord.free.fr/sp

This file is part of Simple Parser.

Simple Parser is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Simple Parser is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with Simple Parser.  If not, see <http://www.gnu.org/licenses/>.
"""
import struct
import sys
import os

import sp
import re
import string

global dot
global ErrorFlag
global AsmFlag
global IncludeFlag
global stopped
global Phase

class Calc:
    def __init__(self):
        self.number = Num
        self.memory = {}
    def __getitem__(self, var): return self.memory[var.name]
    def __setitem__(self, var, val): self.memory[var.name] = val


class Num:
    name, descr = "num", "Number"
    def __init__(self, val):
        if isinstance(val, Num): self.val = val.val
        else: self.val = val
    def __int__(self):      return int(self.val)
    def __str__(self):      return str(self.val)
    def __add__(x, y):      return x.__class__(x.val + y.val)
    def __sub__(x, y):      return x.__class__(x.val - y.val)
    def __or__(x, y):       return x.__class__(x.val | y.val)
    def __xor__(x, y):      return x.__class__(x.val ^ y.val)
    def __mul__(x, y):      return x.__class__(x.val * y.val)
    def __mod__(x, y):      return x.__class__(x.val % y.val)
    def __truediv__(x, y):  return x.__class__(x.val / y.val)
    def __and__(x, y):      return x.__class__(x.val & y.val)
    def __pos__(x):         return x.__class__(+x.val)
    def __neg__(x):         return x.__class__(-x.val)
    def __invert__(x):      return x.__class__(~x.val)

class Bin:
    def __init__(self, n): self.val = int(n.replace('%', ''), 2)
    def eval(self, calc): return calc.number(self.val).val

class Hex:
    def __init__(self, n): self.val = int(n.replace('$', ''), 16)
    def eval(self, calc): return calc.number(self.val).val

class Dec:
    def __init__(self, n): self.val = int(n.replace('_', ''), 10)
    def eval(self, calc): return calc.number(self.val).val


class Var:
    def __init__(self, n): self.name = n
    def eval(self, calc):
        global Phase
        try: val = calc[self]
        except KeyError:
           if Phase ==2:
               raise NameError(self.name)
           else:
               return 3142
        return val #.eval(calc)

class Here:
    global dot
    def __init__(self, n): self.val = dot 
    def eval(self, calc):return calc.number(dot).val

class SetStop:
    global dot
    global stopped
    global IncludeFlag
    def __init__(self, n): self.val = dot 
    def eval(self, calc):
        global dot
        global stopped
        global IncludeFlag
        if IncludeFlag:
            IncludeFlag = False
        else:
            stopped = True
        return ['    ']

class SetStopwXfrAddr:
    global dot
    global stopped
    def __init__(self, expr):self.expr = expr 
    def eval(self, calc):
        global dot
        global stopped
        stopped = True
        XfrAddr  = self.expr.eval(calc) 
        return [ShowHex(XfrAddr)]


class SetHere:
    def __init__(self, expr): self.expr = expr; dot = self.expr 
    def eval(self, calc):
	global dot
        dot  = self.expr.eval(calc) 
        return [ ShowHex(self.expr.eval(calc))]

class AdjustHere:
    def __init__(self, expr):
        global dot
        self.expr = expr
    def eval(self, calc):
	global dot
        olddot = dot
        dot  = dot + self.expr.eval(calc)
        return [ ShowHex(olddot)] 

class CondAsm:
    def __init__(self, expr):
        global Asmflag
        self.expr = expr
    def eval(self, calc):
	global AsmFlag
        val =  self.expr.eval(calc);
        AsmFlag = (val <> 0) 
#        print AsmFlag,val
        if AsmFlag:
            return [ "    " ]
        else: 
            return [ " ..." ]

class Assign:
    def __init__(self, name, expr): self.name, self.expr = name, expr
    def eval(self, calc):
        calc[self.name] = self.expr.eval(calc)
        return [ ShowHex(self.expr.eval(calc))]


class FCB:
    def __init__(self, expr): self.expr =  expr
    def eval(self, calc):
        global dot
        global ErrorFlag
        olddot = dot
        if (self.expr.eval(calc)>256) or (self.expr.eval(calc)<-128) :
             ErrorFlag = '-->T '
        dot = dot + 1
        return [ ShowHex(olddot) ,ShowHex(self.expr.eval(calc),width = 8)]
############################################################
class FCB2:
    def __init__(self, expr1,expr2): self.expr1,self.expr2 =  expr1,expr2
    def eval(self, calc):
        global dot
        global ErrorFlag
        olddot = dot
        data1 = self.expr1.eval(calc)
        if (data1 >256) or (data1<-128) :
             ErrorFlag = '-->T '
        data2 = self.expr2.eval(calc)
        if (data2 >256) or (data2<-128) :
             ErrorFlag = '-->T '
        dot = dot + 2
        return [ ShowHex(olddot) ,ShowHex(data1,width = 8) , #
                                  ShowHex(data2,width = 8) ]
class FCB3:
    def __init__(self, expr1,expr2,expr3): self.expr1,self.expr2,self.expr3 =  expr1,expr2,expr3
    def eval(self, calc):
        global dot
        global ErrorFlag
        olddot = dot
        data1 = self.expr1.eval(calc)
        if (data1 >256) or (data1<-128) :
             ErrorFlag = '-->T '
        data2 = self.expr2.eval(calc)
        if (data2 >256) or (data2<-128) :
             ErrorFlag = '-->T '
        data3 = self.expr3.eval(calc)
        if (data3 >256) or (data3<-128) :
             ErrorFlag = '-->T '
        dot = dot + 3
        return [ ShowHex(olddot) ,ShowHex(data1,width = 8) , #
                                  ShowHex(data2,width = 8) , #
                                  ShowHex(data3,width = 8) ]

class FCB4:
    def __init__(self, expr1,expr2,expr3,expr4): self.expr1,self.expr2,self.expr3,self.expr4 =  expr1,expr2,expr3,expr4
    def eval(self, calc):
        global dot
        global ErrorFlag
        olddot = dot
        data1 = self.expr1.eval(calc)
        if (data1 >256) or (data1<-128) :
             ErrorFlag = '-->T '
        data2 = self.expr2.eval(calc)
        if (data2 >256) or (data2<-128) :
             ErrorFlag = '-->T '
        data3 = self.expr3.eval(calc)
        if (data3 >256) or (data3<-128) :
             ErrorFlag = '-->T '
        data4 = self.expr4.eval(calc)
        if (data4 >256) or (data4<-128) :
             ErrorFlag = '-->T '
        dot = dot + 4
        return [ ShowHex(olddot) ,ShowHex(data1,width = 8) , #
                                  ShowHex(data2,width = 8) , #
                                  ShowHex(data3,width = 8) , #
                                  ShowHex(data4,width = 8) ]


############################################################
class FDB:
    def __init__(self, expr): self.expr = expr
    def eval(self, calc):
        global dot
        global ErrorFlag
        olddot = dot
        if (self.expr.eval(calc)>65536) or (self.expr.eval(calc)<-32768) :
             ErrorFlag = '-->T '
        dot = dot + 2
        return [ ShowHex(olddot) ,ShowHex(self.expr.eval(calc)>>8,width = 8),ShowHex(self.expr.eval(calc),width = 8)]

class FDB2:
    def __init__(self, expr1, expr2): self.expr1, self.expr2=  expr1,expr2
    def eval(self, calc):
        global dot
        global ErrorFlag
        olddot = dot
        data1 =self.expr1.eval(calc)
        data2 =self.expr2.eval(calc)
        if (data1 >65536) or (data1 <-32768) :
             ErrorFlag = '-->T '
        if (data2>65536) or (data2<-32768) :
             ErrorFlag = '-->T '
        dot = dot + 4
        return [ ShowHex(olddot) ,ShowHex(data1 >>8,width = 8),ShowHex(data1 ,width = 8), #
                                  ShowHex(data2 >>8,width = 8),ShowHex(data2 ,width = 8)] 

class InhGroup:
    def __init__(self, opcode): self.opcode = opcode
    def eval(self, calc):
        global dot
        olddot = dot
        dot = dot + 1
        return [ ShowHex(olddot) ,self.opcode]

class RelGroup:
    def __init__(self, opcode, expr): self.opcode, self.expr = opcode, expr
    def eval(self, calc):
        global dot
        global ErrorFlag
        olddot = dot
        rel =  self.expr.eval(calc) - (dot + 2)
        if (rel>127) or (rel<-128) :
             ErrorFlag = '-->T '
        dot = dot + 2
        return [ ShowHex(olddot) ,self.opcode, ShowHex(rel,width = 8)]

class BitSetGroup:
    def __init__(self, opcode, bit, direct): self.opcode, self.bit, self.direct = opcode, bit, direct
    def eval(self, calc):
        global dot
        global ErrorFlag
        olddot = dot
        bit =  self.bit.eval(calc)
        if (bit>7) or (bit<0) :
             ErrorFlag = '-->T '
        if (self.direct.eval(calc)>256) or (self.direct.eval(calc)<0) :
             ErrorFlag = '-->T '
        dot = dot + 2
        return [ ShowHex(olddot) , ShowHex(self.opcode+2*self.bit.eval(calc),width = 8) , ShowHex(self.direct.eval(calc),width = 8)] 

class BitBraGroup:
    def __init__(self, opcode, bit, direct, expr): self.opcode, self.bit, self.direct, self.expr = opcode, bit, direct, expr
    def eval(self, calc):
        global dot
        global ErrorFlag
        olddot = dot
        bit =  self.bit.eval(calc)
        if (bit>7) or (bit<0) :
             ErrorFlag = '-->T '
        if (self.direct.eval(calc)>256) or (self.direct.eval(calc)<0) :
             ErrorFlag = '-->T '
        rel =  self.expr.eval(calc) - (dot + 3)
        if (rel>127) or (rel<-128) :
             ErrorFlag = '-->T '
        dot = dot + 3
        return [ ShowHex(olddot) , ShowHex(self.opcode+2*self.bit.eval(calc),width=8) ,  #
                                   ShowHex(self.direct.eval(calc),width = 8), ShowHex(rel,width = 8)]


class RMWdirGroup:
    def __init__(self, opcode, direct): self.opcode, self.direct = opcode, direct
    def eval(self, calc):
        global dot
        global ErrorFlag
        olddot = dot
        if (self.direct.eval(calc)>256) or (self.direct.eval(calc)<0) :
             ErrorFlag = '-->T '
        dot = dot + 2
        return [ ShowHex(olddot) ,'3'+self.opcode, ShowHex(self.direct.eval(calc),width = 8)]

class RMWix1Group:
    def __init__(self, opcode, expr): self.opcode, self.expr = opcode, expr
    def eval(self, calc):
        global dot
        global ErrorFlag
        olddot = dot
        if (self.expr.eval(calc)>256) or (self.expr.eval(calc)<0) :
             ErrorFlag = '-->T '
        dot = dot + 2
        return [  ShowHex(olddot) ,'6'+self.opcode, ShowHex(self.expr.eval(calc),width = 8)]

class RMWsp1Group:
    def __init__(self, opcode, expr): self.opcode, self.expr = opcode, expr
    def eval(self, calc):
        global dot
        global ErrorFlag
        olddot = dot
        if (self.expr.eval(calc)>256) or (self.expr.eval(calc)<0) :
             ErrorFlag = '-->T '
        dot = dot + 3
        return [ ShowHex(olddot) ,'9E', '6'+self.opcode, ShowHex(self.expr.eval(calc),width = 8)]

class RMWixGroup:
    def __init__(self, opcode): self.opcode = opcode
    def eval(self, calc):
        global dot
        olddot = dot
        dot = dot + 1
        return [ ShowHex(olddot) ,'7'+self.opcode]

class RGMimmGroup:
    def __init__(self, opcode, direct): self.opcode, self.direct = opcode, direct
    def eval(self, calc):
        global dot
        global ErrorFlag
        olddot = dot
        if (self.direct.eval(calc)>255) or (self.direct.eval(calc)<0) :
             ErrorFlag = '-->T '
        dot = dot + 2
        return [ ShowHex(olddot) ,'A'+self.opcode, ShowHex(self.direct.eval(calc),width = 8)]

class RGMdirGroup:
    def __init__(self, opcode, expr): self.opcode, self.expr = opcode, expr
    def eval(self, calc):
        global dot
        global ErrorFlag
        olddot = dot
        data = self.expr.eval(calc)
        if data<0:
            dot = dot + 2
            ErrorFlag = '-->T '
            return [ ShowHex(olddot) ,'B'+self.opcode, ShowHex(data,width=8)]
        elif (data<256) and (data>=0) :
            dot = dot + 2
            return [ ShowHex(olddot) ,'B'+self.opcode, ShowHex(data,width=8)]
        elif data<65536:
            dot = dot + 3
            return [ ShowHex(olddot) ,'C'+self.opcode, ShowHex(data>>8,width=8), ShowHex(data,width=8)]
        else:
            dot = dot + 2
            ErrorFlag = '-->T '
            return [ ShowHex(olddot) ,'B'+self.opcode, ShowHex(data,width=8)]


class RGMix12Group:
    def __init__(self, opcode, expr): self.opcode, self.expr = opcode, expr
    def eval(self, calc):
        global dot
        global ErrorFlag
        olddot = dot
        data = self.expr.eval(calc)
        if data<-128:
            dot = dot + 2
            ErrorFlag = '-->T '
            return [ ShowHex(olddot) ,'E'+self.opcode, ShowHex(data,width=8)]
        elif (data<256) and (data>=-128) :
            dot = dot + 2
            return [ ShowHex(olddot) ,'E'+self.opcode, ShowHex(data,width=8)]
        elif data<65536:
            dot = dot + 3
            return [ ShowHex(olddot) ,'D'+self.opcode, ShowHex(data>>8,width=8), ShowHex(data,width=8)]
        else:
            dot = dot + 2
            ErrorFlag = '-->T '
            return [ ShowHex(olddot) ,'E'+self.opcode, ShowHex(data,width=8)]

class RGMsp12Group:
    def __init__(self, opcode, expr): self.opcode, self.expr = opcode, expr
    def eval(self, calc):
        global dot
        global ErrorFlag
        olddot = dot
        data = self.expr.eval(calc)
        if data<-128:
            dot = dot + 2
            ErrorFlag = '-->T '
            return [ ShowHex(olddot) ,'9E' ,'E'+self.opcode, ShowHex(data,width=8)]
        elif (data<256) and (data>=-128) :
            dot = dot + 2
            return [ ShowHex(olddot) ,'9E' ,'E'+self.opcode, ShowHex(data,width=8)]
        elif data<65536:
            dot = dot + 3
            return [ ShowHex(olddot) ,'9E' ,'D'+self.opcode, ShowHex(data>>8,width=8), ShowHex(data,width=8)]
        else:
            dot = dot + 2
            ErrorFlag = '-->T '
            return [ ShowHex(olddot) ,'9E' ,'E'+self.opcode, ShowHex(data,width=8)]

class RGMixGroup:
    def __init__(self, opcode): self.opcode = opcode
    def eval(self, calc):
        global dot
        olddot = dot
        dot = dot + 1
        return [ ShowHex(olddot) ,'F'+self.opcode]

class LDHXdirGroup:
    def __init__(self, opcode, expr): self.opcode, self.expr = opcode, expr
    def eval(self, calc):
        global dot
        global ErrorFlag
        olddot = dot
        data = self.expr.eval(calc)
        if data<0:
            dot = dot + 2
            ErrorFlag = '-->T '
            return [ ShowHex(olddot) ,'55', ShowHex(data,width=8)]
        elif (data<256) and (data>=0) :
            dot = dot + 2
            return [ ShowHex(olddot) ,'55', ShowHex(data,width=8)]
        elif data<65536:
            dot = dot + 3
            return [ ShowHex(olddot) ,'32', ShowHex(data>>8,width=8), ShowHex(data,width=8)]
        else:
            dot = dot + 2
            ErrorFlag = '-->T '
            return [ ShowHex(olddot) ,'55', ShowHex(data,width=8)]

class LDHXimmGroup:
    def __init__(self, opcode, imm): self.opcode, self.imm = opcode, imm
    def eval(self, calc):
        global dot
        global ErrorFlag
        olddot = dot
        data = self.imm.eval(calc)
        if (data>65536) or (data<-32768) :
             ErrorFlag = '-->T '
        dot = dot + 3
        return [ ShowHex(olddot) ,'45', ShowHex(data>>8,width=8), ShowHex(data,width=8)]

class LDHXix12Group:
    def __init__(self, opcode, expr): self.opcode, self.expr = opcode, expr
    def eval(self, calc):
        global dot
        global ErrorFlag
        olddot = dot
        data = self.expr.eval(calc)
        if data<-128:
            dot = dot + 3
            ErrorFlag = '-->T '
            return [ ShowHex(olddot) , '9E', 'CE', ShowHex(data,width=8)]
        elif (data<256) and (data>=-128) :
            dot = dot + 3
            return [ ShowHex(olddot) , '9E', 'CE', ShowHex(data,width=8)]
        elif data<65536:
            dot = dot + 3
            return [ ShowHex(olddot) , '9E', 'BE', ShowHex(data>>8,width=8), ShowHex(data,width=8)]
        else:
            dot = dot + 3
            ErrorFlag = '-->T '
            return [ ShowHex(olddot) , '9E', 'CE', ShowHex(data,width=8)]

class LDHXsp1Group:
    def __init__(self, opcode, expr): self.opcode, self.expr = opcode, expr
    def eval(self, calc):
        global dot
        global ErrorFlag
        olddot = dot
        data = self.expr.eval(calc)
        if (data<256) and (data>=-128) :
            dot = dot + 2
            return [ ShowHex(olddot) , '9E', 'FE', ShowHex(data,width=8)]
        else:
            dot = dot + 2
            ErrorFlag = '-->T '
            return [ ShowHex(olddot) , '9E', 'FE', ShowHex(data,width=8)]

class LDHXixGroup:
    def __init__(self, opcode): self.opcode = opcode
    def eval(self, calc):
        global dot
        olddot = dot
        dot = dot + 2
        return [ ShowHex(olddot) , '9E', 'AE']

class CPHXimm:
    def __init__(self, opcode, imm): self.opcode, self.imm = opcode, imm
    def eval(self, calc):
        global dot
        global ErrorFlag
        olddot = dot
        data = self.imm.eval(calc)
        if (data>65536) or (data<-32768) :
             ErrorFlag = '-->T '
        dot = dot + 3
        return [ ShowHex(olddot) ,'65', ShowHex(data>>8,width=8), ShowHex(data,width=8)]

class CPHXdirGroup:
    def __init__(self, opcode, expr): self.opcode, self.expr = opcode, expr
    def eval(self, calc):
        global dot
        global ErrorFlag
        olddot = dot
        data = self.expr.eval(calc)
        if data<0:
            dot = dot + 2
            ErrorFlag = '-->T '
            return [ ShowHex(olddot) ,'75', ShowHex(data,width=8)]
        elif (data<256) and (data>=0) :
            dot = dot + 2
            return [ ShowHex(olddot) ,'75', ShowHex(data,width=8)]
        elif data<65536:
            dot = dot + 3
            return [ ShowHex(olddot) ,'3E', ShowHex(data>>8,width=8), ShowHex(data,width=8)]
        else:
            dot = dot + 2
            ErrorFlag = '-->T '
            return [ ShowHex(olddot) ,'75', ShowHex(data,width=8)]

class CPHXsp1Group:
    def __init__(self, opcode, expr): self.opcode, self.expr = opcode, expr
    def eval(self, calc):
        global dot
        global ErrorFlag
        olddot = dot
        data = self.expr.eval(calc)
        if (data<256) and (data>=-128) :
            dot = dot + 2
            return [ ShowHex(olddot) , '9E', 'F3', ShowHex(data,width=8)]
        else:
            dot = dot + 2
            ErrorFlag = '-->T '
            return [ ShowHex(olddot) , '9E', 'F3', ShowHex(data,width=8)]

class STHXsp1:
    def __init__(self, opcode, expr): self.opcode, self.expr = opcode, expr
    def eval(self, calc):
        global dot
        global ErrorFlag
        olddot = dot
        data = self.expr.eval(calc)
        if (data<256) and (data>=-128) :
            dot = dot + 2
            return [ ShowHex(olddot) , '9E', 'FF', ShowHex(data,width=8)]
        else:
            dot = dot + 2
            ErrorFlag = '-->T '
            return [ ShowHex(olddot) , '9E', 'FF', ShowHex(data,width=8)]

class STHXdir:
    def __init__(self, opcode, expr): self.opcode, self.expr = opcode, expr
    def eval(self, calc):
        global dot
        global ErrorFlag
        olddot = dot
        data = self.expr.eval(calc)
        if (data<256) and (data>=0) :
            dot = dot + 2
            return [ ShowHex(olddot) ,'35', ShowHex(data,width=8)]
        else:
            dot = dot + 2
            ErrorFlag = '-->T '
            return [ ShowHex(olddot) ,'35', ShowHex(data,width=8)]

class CBEQAimmGroup:
    def __init__(self, opcode, direct, expr): self.opcode, self.direct, self.expr = opcode, direct, expr
    def eval(self, calc):
        global dot
        global ErrorFlag
        olddot = dot
        if (self.direct.eval(calc)>256) or (self.direct.eval(calc)<0) :
             ErrorFlag = '-->T '
        rel =  self.expr.eval(calc) - (dot + 3)
        if (rel>127) or (rel<-128) :
             ErrorFlag = '-->T '
        dot = dot + 3
        return [ ShowHex(olddot) , self.opcode ,  #
                                   ShowHex(self.direct.eval(calc),width = 8), ShowHex(rel,width = 8)]

class CBEQdir:
    def __init__(self, opcode, direct, expr): self.opcode, self.direct, self.expr = opcode, direct, expr
    def eval(self, calc):
        global dot
        global ErrorFlag
        olddot = dot
        if (self.direct.eval(calc)>256) or (self.direct.eval(calc)<0) :
             ErrorFlag = '-->T '
        rel =  self.expr.eval(calc) - (dot + 3)
        if (rel>127) or (rel<-128) :
             ErrorFlag = '-->T '
        dot = dot + 3
        return [ ShowHex(olddot) , '31' ,ShowHex(self.direct.eval(calc),width = 8), ShowHex(rel,width = 8)]

class DBNZdir:
    def __init__(self, opcode, direct, expr): self.opcode, self.direct, self.expr = opcode, direct, expr
    def eval(self, calc):
        global dot
        global ErrorFlag
        olddot = dot
        if (self.direct.eval(calc)>256) or (self.direct.eval(calc)<0) :
             ErrorFlag = '-->T '
        rel =  self.expr.eval(calc) - (dot + 3)
        if (rel>127) or (rel<-128) :
             ErrorFlag = '-->T '
        dot = dot + 3
        return [ ShowHex(olddot) , '3B' ,ShowHex(self.direct.eval(calc),width = 8), ShowHex(rel,width = 8)]

class CBEQix1:
    def __init__(self, opcode, direct, expr): self.opcode, self.direct, self.expr = opcode, direct, expr
    def eval(self, calc):
        global dot
        global ErrorFlag
        olddot = dot
        if (self.direct.eval(calc)>256) or (self.direct.eval(calc)<0) :
             ErrorFlag = '-->T '
        rel =  self.expr.eval(calc) - (dot + 3)
        if (rel>127) or (rel<-128) :
             ErrorFlag = '-->T '
        dot = dot + 3
        return [ ShowHex(olddot) , '61' ,ShowHex(self.direct.eval(calc),width = 8), ShowHex(rel,width = 8)]

class DBNZix1:
    def __init__(self, opcode, direct, expr): self.opcode, self.direct, self.expr = opcode, direct, expr
    def eval(self, calc):
        global dot
        global ErrorFlag
        olddot = dot
        if (self.direct.eval(calc)>256) or (self.direct.eval(calc)<0) :
             ErrorFlag = '-->T '
        rel =  self.expr.eval(calc) - (dot + 3)
        if (rel>127) or (rel<-128) :
             ErrorFlag = '-->T '
        dot = dot + 3
        return [ ShowHex(olddot) , '6B' ,ShowHex(self.direct.eval(calc),width = 8), ShowHex(rel,width = 8)]

class CBEQix:
    def __init__(self, opcode, expr): self.opcode, self.expr = opcode, expr
    def eval(self, calc):
        global dot
        global ErrorFlag
        olddot = dot
        rel =  self.expr.eval(calc) - (dot + 2)
        if (rel>127) or (rel<-128) :
             ErrorFlag = '-->T '
        dot = dot + 2
        return [ ShowHex(olddot) ,'71', ShowHex(rel,width = 8)]

class DBNZix:
    def __init__(self, opcode, expr): self.opcode, self.expr = opcode, expr
    def eval(self, calc):
        global dot
        global ErrorFlag
        olddot = dot
        rel =  self.expr.eval(calc) - (dot + 2)
        if (rel>127) or (rel<-128) :
             ErrorFlag = '-->T '
        dot = dot + 2
        return [ ShowHex(olddot) ,'7B', ShowHex(rel,width = 8)]

#######################################################################

class MOVdir:
    def __init__(self, expr1,expr2): self.expr1,self.expr2 =  expr1,expr2
    def eval(self, calc):
        global dot
        global ErrorFlag
        olddot = dot
        data1 = self.expr1.eval(calc)
        if (data1 >256) or (data1<0) :
             ErrorFlag = '-->T '
        data2 = self.expr2.eval(calc)
        if (data2 >256) or (data2<0) :
             ErrorFlag = '-->T '
        dot = dot + 3
        return [ ShowHex(olddot) , '4E' ,ShowHex(data1,width = 8) ,ShowHex(data2,width = 8) ]

class MOVimm:
    def __init__(self, expr1,expr2): self.expr1,self.expr2 =  expr1,expr2
    def eval(self, calc):
        global dot
        global ErrorFlag
        olddot = dot
        data1 = self.expr1.eval(calc)
        if (data1 >256) or (data1<-128) :
             ErrorFlag = '-->T '
        data2 = self.expr2.eval(calc)
        if (data2 >256) or (data2<0) :
             ErrorFlag = '-->T '
        dot = dot + 3
        return [ ShowHex(olddot) , '6E' ,ShowHex(data1,width = 8) ,ShowHex(data2,width = 8) ]

class MOVdix:
    def __init__(self, expr): self.expr =  expr
    def eval(self, calc):
        global dot
        global ErrorFlag
        olddot = dot
        if (self.expr.eval(calc)>256) or (self.expr.eval(calc)<0) :
             ErrorFlag = '-->T '
        dot = dot + 2
        return [ ShowHex(olddot) , '5E', ShowHex(self.expr.eval(calc),width = 8)]

class MOVixd:
    def __init__(self, expr): self.expr =  expr
    def eval(self, calc):
        global dot
        global ErrorFlag
        olddot = dot
        if (self.expr.eval(calc)>256) or (self.expr.eval(calc)<0) :
             ErrorFlag = '-->T '
        dot = dot + 2
        return [ ShowHex(olddot) , '7E', ShowHex(self.expr.eval(calc),width = 8)]


#######################################################################
class Op2:
    def __init__(self, x, y): self.x, self.y = x, y

class Op1:
    def __init__(self, x): self.x = x


class Add(Op2):
    def eval(self, calc): return self.x.eval(calc) + self.y.eval(calc)

class Sub(Op2):
    def eval(self, calc): return self.x.eval(calc) - self.y.eval(calc)

class Or(Op2):
    def eval(self, calc): return self.x.eval(calc) | self.y.eval(calc)

class Xor(Op2):
    def eval(self, calc): return self.x.eval(calc) ^ self.y.eval(calc)

class Mul(Op2):
    def eval(self, calc): return self.x.eval(calc) * self.y.eval(calc)

class Div(Op2):
    def eval(self, calc): return self.x.eval(calc) / self.y.eval(calc)

class And(Op2):
    def eval(self, calc): return self.x.eval(calc) & self.y.eval(calc)

class Pos(Op1):
    def eval(self, calc): return +self.x.eval(calc)

class Neg(Op1):
    def eval(self, calc): return -self.x.eval(calc)

class Inv(Op1):
    def eval(self, calc): return ~self.x.eval(calc)

class BitPos(Op1):
    def eval(self, calc):
        ln={ 1:0 , 2:1, 4:2, 8:3, 16:4, 32:5, 64:6, 128:7}
        return ln.get(self.x.eval(calc),-1)

def red(x, ys):
    for f, y in ys:
        x = f(x, y)
    return x

def red1(f, x):
    return f(x)

def ShowHex(N, width=16):
    if width:
        N %= 2**width
        bits_per_digit = 4
        min_len = width//bits_per_digit
#    if N < 0:
#        N = -N
#        sign = "-"
#    else:
#        sign = ""
    s = ""
    while N: #base
        N, d = divmod(N, 16)
        s = s + "0123456789ABCDEF"[d]
    s = s or "0"
    if width:
        s = s + "0"*(min_len-len(s))
#    s = " ".join(s[i:i+group] for i in range(0, len(s), group))
    return s[::-1]

Parser = sp.compile(
    r"""
        dec = r'[0-9]+' : `Dec` ;
        bin = r'%([0-1]+)\b' : `Bin` ;
        hex = r'[$]([0-9a-fA-F]+)\b' : `Hex` ;
        var = r'[a-zA-Z_]\w*' : `Var`;
        here = '.' : `Here` ;

        equ = (' EQU '|' equ ') ;
        org = ('ORG '|'org ') ;
        condasm = ('IF '|'if ') ;
        rmb = ('RMB '|'rmb ') ;
        fcb = ('FCB '|'fcb ') ;
        fdb = ('FDB '|'fdb ') ;
        end = ('END'|'end') ;

        rolx = ('NEGA'|'nega') `'40'` ;
        rolx = ('MUL'|'mul')   `'42'` ;
        rolx = ('COMA'|'coma') `'43'` ;
        rolx = ('LSRA'|'lsra') `'44'` ;
        rolx = ('ROLA'|'rola') `'46'` ;
        rolx = ('ASRA'|'asra') `'47'` ;
        rolx = ('LSLA'|'lsla') `'48'` ;
        rolx = ('ASLA'|'asla') `'48'` ;
        rolx = ('ROLA'|'rola') `'49'` ;
        rolx = ('DECA'|'deca') `'4A'` ;
        rolx = ('INCA'|'inca') `'4C'` ;
        rolx = ('TSTA'|'tsta')  `'4C'` ;
        rolx = ('CLRA'|'clra') `'4D'` ;
        rolx = ('NEGX'|'negx') `'5C'` ;
        rolx = ('DIV'|'div')   `'52'` ;
        rolx = ('COMX'|'comx') `'53'` ;
        rolx = ('LSRX'|'lsrx') `'54'` ;
        rolx = ('RORX'|'rorx') `'57'` ;
        rolx = ('ASRX'|'asrx') `'5C'` ;
        rolx = ('LSLX'|'lslx') `'58'` ;
        rolx = ('ASLX'|'aslx') `'58'` ;
        rolx = ('ROLX'|'rolx') `'59'` ;
        rolx = ('DECX'|'decx') `'5A'` ;
        rolx = ('INCX'|'incx') `'5D'` ;
        rolx = ('TSTX'|'tstx') `'5C'` ;
        rolx = ('CLRX'|'clrx') `'5F'` ;
        rolx = ('NSA'|'nsa')   `'62'` ;
        rolx = ('DAA'|'daa')   `'72'` ;
        rolx = ('RTI'|'rti')   `'80'` ;
        rolx = ('RTS'|'rts')   `'81'` ;
        rolx = ('BGND'|'bgnd') `'82'` ;
        rolx = ('SWI'|'swi')   `'83'` ;
        rolx = ('TAP'|'tap')   `'84'` ;
        rolx = ('TPA'|'tpa')   `'85'` ;
        rolx = ('PULA'|'pula') `'86'` ;
        rolx = ('PSHA'|'psha') `'87'` ;
        rolx = ('PULX'|'pulx') `'88'` ;
        rolx = ('PSHX'|'pshx') `'89'` ;
        rolx = ('PULH'|'pulh') `'8A'` ;
        rolx = ('PSHH'|'pshh') `'8B'` ;
        rolx = ('CLRH'|'clrh') `'8C'` ;
        rolx = ('STOP'|'stop') `'8E'` ;
        rolx = ('WAI'|'wai')   `'8F'` ;
        rolx = ('TXS'|'txs')   `'94'` ;
        rolx = ('TSX'|'tsx')   `'95'` ;
        rolx = ('TAX'|'tax')   `'97'` ;
        rolx = ('CLC'|'clc')   `'98'` ;
        rolx = ('SEC'|'sec')   `'99'` ;
        rolx = ('CLI'|'cli')   `'9A'` ;
        rolx = ('SEI'|'sei')   `'9B'` ;
        rolx = ('RSP'|'rsp')   `'9C'` ;
        rolx = ('NOP'|'nop')   `'9D'` ;
        rolx = ('TXA'|'txa')   `'9F'` ;



        brset = ('BRSET '|'brset ') `0` ;
        brset = ('BRCLR '|'brclr ') `1` ;
        bset = ('BSET '|'bset ') `16` ;
        bset = ('BCLR '|'bclr ') `17` ;

        bra = ('BRA '|'bra ') `'20'` ;
        bra = ('BRN '|'brn ') `'21'` ;
        bra = ('BHI '|'bhi ') `'22'` ;
        bra = ('BHS '|'bhs ') `'24'` ;
        bra = ('BLS '|'bls ') `'23'` ;
        bra = ('BCC '|'bcc ') `'24'` ;
        bra = ('BCS '|'bcs ') `'25'` ;
        bra = ('BNE '|'bne ') `'26'` ;
        bra = ('BEQ '|'beq ') `'27'` ;
        bra = ('BHCC '|'bhcc ') `'28'` ;
        bra = ('BHCS '|'bhcs ') `'29'` ;
        bra = ('BPL '|'bpl ') `'2A'` ;
        bra = ('BMI '|'bmi ') `'2B'` ;
        bra = ('BMC '|'bmc ') `'2C'` ;
        bra = ('BMS '|'bms ') `'2D'` ;
        bra = ('BIL '|'bil ') `'2E'` ;
        bra = ('BIH '|'bih ') `'2F'` ;
        bra = ('BGE '|'bge ') `'90'` ;
        bra = ('BLT '|'blt ') `'91'` ;
        bra = ('BGT '|'bgt ') `'92'` ;
        bra = ('BLE '|'ble ') `'93'` ;
        bra = ('BLO '|'blo ') `'25'` ;
        bra = ('BSR '|'bsr ') `'AD'` ;
        bra = ('DBNZA '|'dbnza ') `'4B'` ;
        bra = ('DBNZX '|'dbnzx ') `'5B'` ;

        tst = ('NEG '| 'neg ') `'0'` ;
        tst = ('COM '| 'com ') `'3'` ;
        tst = ('LSR '| 'lsr ') `'4'` ;
        tst = ('ROR '| 'ror ') `'6'` ;
        tst = ('ASR '| 'asr ') `'7'` ;
        tst = ('LSL '| 'lsl ') `'8'` ;
        tst = ('ASL '| 'asl ') `'8'` ;
        tst = ('ROL '| 'rol ') `'9'` ;
        tst = ('DEC '| 'dec ') `'A'` ;
        tst = ('INC '| 'inc ') `'C'` ;
        tst = ('TST '| 'tst ') `'D'` ;
        tst = ('CLR '| 'clr ') `'F'` ;
 
        adc = ('SUB '|'sub ')  `'0'` ;
        adc = ('CMP '|'cmp ')  `'1'` ;
        adc = ('SBC '|'sbc ')  `'2'` ;
        adc = ('CPX '|'cpx ')  `'3'` ;
        adc = ('AND '|'and ')  `'4'` ;
        adc = ('BIT '|'bit ')  `'5'` ;
        adc = ('LDA '|'lda ')  `'6'` ;
        adc = ('EOR '|'eor ')  `'8'` ;
        adc = ('ADC '|'adc ')  `'9'` ;
        adc = ('ORA '|'ora ')  `'A'` ;
        adc = ('ADD '|'add ')  `'B'` ;
        adc = ('LDX '|'ldx ')  `'E'` ;


        sta = ('STA '|'sta ') `'7'` ;
        sta = ('STX '|'stx ') `'F'` ;

        jmp = ('JMP '|'jmp ') `'C'` ;
        jmp = ('JSR '|'jsr ') `'D'` ;
       
        aix = ('aix '|'AIX ') `'F'` ;
        aix = ('ais '|'AIS ') `'7'` ;

        cphx = ('CPHX '|'cphx ') `'@'` ;

        ldhx = ('LDHX '|'ldhx ') `'@'` ;

        sthx = ('STHX '|'sthx ') `'@'` ;

        mov = ('MOV '|'mov ')  ;

        cbeqa = ('CBEQA '|'cbeqa ') `'41'` ;
        cbeqa = ('CBEQX '|'cbeqx ') `'51'` ;

        cbeq  = ('CBEQ '|'cbeq ') `'1'` ;

        dbnz = ('DBNZ '|'dbnz ') `'7'` ;

        addop = '+' `Add` ;
        addop = '-' `Sub` ;
        addop = '|' `Or` ;
        addop = '^' `Xor` ;  
        mulop = '*' `Mul` ;
        mulop = '/' `Div` ;
        mulop = '&' `And` ;

        unop = '+' `Pos` ;
        unop = '-' `Neg` ;
        unop = '~' `Inv` ;
        unop = '[' `BitPos` ;


        !S = org expr : `SetHere` ;
        !S = rmb expr : `AdjustHere` ;
        !S = condasm expr : `CondAsm` ;
        !S = fcb expr : `FCB`;
        !S = fcb expr "," expr :: `FCB2`;
        !S = fcb expr "," expr "," expr :: `FCB3`;
        !S = fcb expr "," expr "," expr "," expr :: `FCB4`;

        !S = fdb expr : `FDB` ;
        !S = fdb expr "," expr :: `FDB2`;

        !S = end : `SetStop`;
        !S = rolx : `InhGroup`;

        !S = var equ expr :: `Assign` ;

        !S = bra expr :: `RelGroup` ;

        !S = adc '#' expr :: `RGMimmGroup` ;
        !S = adc (',x'|',X') :: `RGMixGroup` ;
        !S = adc expr (',x'|',X') :: `RGMix12Group`;
        !S = adc expr (',sp'|',SP') :: `RGMsp12Group`  ;
        !S = adc expr :: `RGMdirGroup`;

        !S = tst (',x'|',X') :: `RMWixGroup` ;
        !S = tst expr (',x'|',X') :: `RMWix1Group` ;
        !S = tst expr (',sp'|',SP'):: `RMWsp1Group`  ;
        !S = tst expr :: `RMWdirGroup` ;

        !S = sta (',x'|',X') : `RGMixGroup` ;
        !S = sta expr (',x'|',X') :: `RGMix12Group` ;
        !S = sta expr (',sp'|',SP') :: `RGMsp12Group` ;
        !S = sta expr :: `RGMdirGroup` ;

        !S = jmp (',x'|',X') : `RGMixGroup` ;
        !S = jmp expr (',x'|',X') :: `RGMix12Group` ;
        !S = jmp expr :: `RGMdirGroup` ;

        !S = aix '#' expr :: `RGMimmGroup` ;

        !S = brset expr ',' expr ',' expr :: `BitBraGroup`  ;
        !S = bset expr ',' expr :: `BitSetGroup` ; 

        !S = ldhx '#' expr :: `LDHXimmGroup` ;
        !S = ldhx (',x'|',X') :: `LDHXixGroup` ;
        !S = ldhx expr (',x'|',X') :: `LDHXix12Group`;
        !S = ldhx expr (',sp'|',SP') :: `LDHXsp1Group`  ;
        !S = ldhx expr :: `LDHXdirGroup`;

        !S = cphx  '#' expr :: `CPHXimm`;
        !S = cphx  expr (',SP' | ',sp') :: `CPHXsp1Group`;
        !S = cphx  expr :: `CPHXdirGroup`;

        !S = sthx expr (',SP' | ',sp') :: `STHXsp1`;
        !S = sthx expr :: `STHXdir`   ;

        !S = mov '#' expr ',' expr ::`MOVimm`;
        !S = mov (',X+,'|',x+,') expr : `MOVixd`;
        !S = mov expr (',X+'|',x+') : `MOVdix` ;
        !S = mov expr ',' expr :: `MOVdir`;

        !S = cbeqa '#' expr ',' expr :: `CBEQAimmGroup` ;

        !S = cbeq (',X+,'|',x+,') expr :: `CBEQix` ;
        !S = cbeq expr (',X+,'|',x+,') expr :: `CBEQix1` ;
        !S = cbeq expr ',' expr :: `CBEQdir`;

        !S = dbnz (',X,'|',x,') expr:: `DBNZix`;
        !S = dbnz expr  (',X,'|',x,') expr :: `DBNZix1`;
        !S = dbnz expr ',' expr :: `DBNZdir`;


        !S = expr;

        expr = term (addop term)* :: `red` ;
        term = fact (mulop fact)* :: `red` ;
        fact = unop fact :: `red1` | atom ;

        atom = '(' expr ')' ;
        atom = bin | hex | dec | var | here ;

    """)


##        !S = expr;


## this is a debug statement in parser         !S = expr ;
## not clear how to comment it out without
## moving it out of """
## !S = expr ; is necessary for double def to work  need to fix lines that just contain expression

#parser(expr).eval(calc)
#calc = Calc()
#    print(Help().eval())
#    print()

#print("*"*70)
#print sys.argv
source = sys.argv[1]
basefilename = os.path.splitext(source)
#print basefilename
#CantAssembleFlag = False
try:
  src = open(source) #'mactst.asm')
except Exception as e :
    print("Error: No such file "+source)
#    CantAssembleFlag = True 
#if CantAssembleFlag == False:
else:
  listing = open(basefilename[0]+'.lst',"w")
  wholeline = ("#"*10+' '+source+' '+"#"*40)
  print wholeline
  listing.write(wholeline + '\n')

  if __name__ == '__main__':

    calc = Calc()


################################################### FIRST PASS
    Phase =1
      

    inc = open('HC9S08JS8.hdr')
    IncludeFlag = True
    stopped = False
    dot =0
    AsmFlag = True
############################### macro processor, text substitution
    MacroPos = -1
    MacroName =""
    MacroBody=[]
    MacroDict = {}
#################################################################
    while not stopped:
#        stopped = True
#$#$#$#$#$#$#$#$#$#$#$#$#$#$#$#$#$#$#$#$#$#$#$#$#
#################################################################
        if MacroPos <> -1 :
            nextline = MacroBody[MacroPos]   ########### do argument substitutions here
            for x in range(len(MacroArgs)):
               nextline = nextline.replace("@"+str(x),MacroArgs[x])
            MacroPos = MacroPos + 1
        else:
###### this is the only read that matters for include
            if IncludeFlag:
                nextline = inc.readline()[:-1] # no endof line char
            else:
                nextline = string.expandtabs(src.readline()[:-1],1) # no endof line char
                if len(nextline) == 0:
                    nextline =" END ; Unexpected EOF"
########################################################################
        if MacroName<>"":
           MacroBody.append(nextline)
           if nextline.find("mend")<>-1 or nextline.find("MEND")<>-1:
              MacroDict[MacroName] = MacroBody
              MacroName = ""
              AsmFlag = True
              nextline = ';'+nextline             
        else:
            comment = string.split(nextline, ";", 1) # remove comment
            assemble = string.lstrip(re.sub("\s+", " ", string.rstrip(comment[0]))) # crush whitespace
            expand = string.split(assemble," ",1)
            if MacroDict.has_key(expand[0]):
                MacroBody = MacroDict[expand[0]]
                if len(expand) == 2:
                    MacroArgs = eval("['"+ShowHex(dot)+"','"+expand[1]+"']")
                else:
                    MacroArgs = eval("['"+ShowHex(dot)+"']")
                MacroPos = 0
                nextline = ';'+nextline
            if assemble.find("macro") <> -1 or assemble.find("MACRO") <> -1 :
                MacroName = string.split(assemble," ",1)[1]
                MacroBody = []
                AsmFlag = False           
            if assemble.find("mend")<>-1 or assemble.find("MEND")<>-1:
                nextline = ';'+nextline
                MacroPos = -1
#            if assemble.startswith("end"):
#               stopped = True
#        print nextline
################################

#$#$#$#$#$#$#$#$#$#$#$#$#$#$#$#$#$#$#$#$#$#$#$#$#
#$#$#$#$#$#$#$#$#$        nextline = src.readline()[:-1] # no endof line char
        comment = string.split(nextline, ";", 1) # remove comment
        if string.lstrip(comment[0]) <>"":
            if comment[0].startswith(" "):
                assemble = string.lstrip(re.sub("\s+", " ", string.rstrip(comment[0]))) # crush whitespace
            else:
                labeled = string.split(comment[0], " ", 1)
                if  string.lstrip(labeled[1]).upper().startswith("EQU "):
                    assemble = string.lstrip(re.sub("\s+", " ", string.rstrip(comment[0]))) # crush whitespace
                else:
                    if AsmFlag:
#                        print labeled
                        val = Parser(labeled[0]+" equ .").eval(calc)
                    assemble = string.lstrip(re.sub("\s+", " ", string.rstrip(labeled[1]))) # crush whitespace
            ErrorFlag ='     '
            if  AsmFlag or (assemble.find("IF ")<>-1) or (assemble.find("if ")<>-1):
                try:
                    value = Parser(assemble).eval(calc)
                except Exception as e:
                    if e.__class__.__name__ == "NameError":
                       pass #print(" "*16+"-->U "+nextline)
                    elif e.__class__.__name__ == "SyntaxError":
                       pass #print(" "*16+"-->S "+nextline)
                    else:
                       pass #print("%s: %s"%(e.__class__.__name__, e))
                else:
                    try:
                        if len(value) == 1:
                           pass #print(value[0]+("   "*4)+ErrorFlag+nextline) 
                        elif len(value) == 2:
                           pass #print(value[0]+' '+value[1]+("   "*3)+ErrorFlag+nextline) 
                        elif len(value) == 3:
                           pass #print(value[0]+' '+value[1]+' '+value[2]+("   "*2)+ErrorFlag+nextline) 
                        elif len(value) == 4:
                           pass #print(value[0]+' '+value[1]+' '+value[2]+' '+value[3]+("   "*1)+ErrorFlag+nextline) 
                        elif len(value) == 5:
                           pass #print(value[0]+' '+value[1]+' '+value[2]+' '+value[3]+' '+value[4]+ErrorFlag+nextline)
                    except:
                         pass
            else:
               pass #print(" "*21+nextline)
###########################
        else:
           pass #print(" "*21+nextline)
    src.close()
##################################################### SECOND PASS
    Phase =2
    src = open(source) #'mactst.asm')
    S019 = open(basefilename[0]+'.S19',"w")
    stopped = False
    dot =0
    AsmFlag = True
    ErrorCount = 0
############################### macro processor, text substitution
    MacroPos = -1
    MacroName =""
    MacroBody=[]
    MacroDict = {}
#################################################################

    while not stopped:
#        stopped = True
        ErrorFlag ='     '
#$#$#$#$#$#$#$#$#$#$#$#$#$#$#$#$#$#$#$#$#$#$#$#$#
#################################################################
        MustAsmFlag = True
        ExpandFlag = " "
        if MacroPos <> -1 :
            nextline = MacroBody[MacroPos]   ########### do argument substitutions here
            for x in range(len(MacroArgs)):
               nextline = nextline.replace("@"+str(x),MacroArgs[x])
            MacroPos = MacroPos + 1
##### nextline here is from a macro substitution @#@#@#@#@#@#@#@#@#@#@#@#@#@#@#@#@#@#@##@#@#@##@
            ExpandFlag = "|"
        else:
            nextline = string.expandtabs(src.readline()[:-1],1) # no endof line char
            if len(nextline) == 0:
               nextline =" END ; Unexpected EOF"
               ErrorFlag = "-->"
        if MacroName<>"":
           MacroBody.append(nextline)
           if nextline.find("mend")<>-1 or nextline.find("MEND")<>-1:
              MacroDict[MacroName] = MacroBody
              MacroName = ""
              AsmFlag = True
              MustAsmFlag = False 
           if (nextline.find("IF ")<> 0) or (nextline.find("if ")<>0):
              MustAsmFlag = False            
        else:
            comment = string.split(nextline, ";", 1) # remove comment
            assemble = string.lstrip(re.sub("\s+", " ", string.rstrip(comment[0]))) # crush whitespace
            expand = string.split(assemble," ",1)
            if MacroDict.has_key(expand[0]):
                MacroBody = MacroDict[expand[0]]
                if len(expand) == 2:
                    MacroArgs = eval("['"+ShowHex(dot)+"','"+expand[1]+"']")
                else:
                    MacroArgs = eval("['"+ShowHex(dot)+"']")
                MacroPos = 0
                MustAsmFlag = False #nextline = ';'+nextline
            if assemble.find("macro") <> -1 or  assemble.find("MACRO") <> -1:
                MacroName = string.split(assemble," ",1)[1]
                MacroBody = []
                AsmFlag = False           
            if assemble.find("mend")<>-1 or assemble.find("MEND")<>-1:
                MustAsmFlag = False #nextline = ';'+nextline
                MacroPos = -1
            if assemble.startswith("end") or assemble.startswith("END") :
               stopped = True
#        print nextline
################################

#$#$#$#$#$#$#$#$#$#$#$#$#$#$#$#$#$#$#$#$#$#$#$#$#
#$#$#$#$#$#$#        nextline = src.readline()[:-1] # no endof line char
        comment = string.split(nextline, ";", 1) # remove comment
        if string.lstrip(comment[0]) <>"":
            if comment[0].startswith(" "):
                assemble = string.lstrip(re.sub("\s+", " ", string.rstrip(comment[0]))) # crush whitespace
            else:
                labeled = string.split(comment[0], " ", 1)
                if  string.lstrip(labeled[1]).upper().startswith("EQU ."):
                    if AsmFlag:
#                        pass
                       val = Parser(labeled[0]).eval(calc)
#                       print (labeled[0],val,dot)
                       if val <> dot:
                            ErrorFlag = '-->D '
                    assemble = string.lstrip(re.sub("\s+", " ", string.rstrip(comment[0]))) # crush whitespace
                elif  string.lstrip(labeled[1]).upper().startswith("EQU "):
                    assemble = string.lstrip(re.sub("\s+", " ", string.rstrip(comment[0]))) # crush whitespace
                else:
                    if AsmFlag:
                       val = Parser(labeled[0]).eval(calc)
                       if val <> dot:
                            ErrorFlag = '-->D '
                    assemble = string.lstrip(re.sub("\s+", " ", string.rstrip(labeled[1]))) # crush whitespace
##########################################
            if  MustAsmFlag and (AsmFlag or (assemble.find("IF ")<>-1) or (assemble.find("if ")<>-1)):
                try:
                    value = Parser(assemble).eval(calc)
                except Exception as e:
                    if e.__class__.__name__ == "NameError":
                       ErrorFlag ="-->U "
                       print(" "*16+ErrorFlag+ExpandFlag+nextline)
                    elif e.__class__.__name__ == "SyntaxError":
                       ErrorFlag ="-->S "
                       print(" "*16+ErrorFlag+ExpandFlag+nextline)
                    else:
                       print("%s: %s"%(e.__class__.__name__, e))
                else:
                    if len(value) == 1:
                       wholeline = (value[0]+("   "*4)) 
                    elif len(value) == 2:
                       chksum = 4+ int(value[0][:2],16)+int(value[0][2:],16)+int(value[1],16)
                       wholeline = (value[0]+' '+value[1]+("   "*3))
                       S019.write("S104"+string.replace(wholeline," ","")+ShowHex(~chksum,8)+'\n') 
                    elif len(value) == 3:
                       chksum = 5+ int(value[0][:2],16)+int(value[0][2:],16)+int(value[1],16)+int(value[2],16)
                       wholeline = (value[0]+' '+value[1]+' '+value[2]+("   "*2)) 
                       S019.write("S105"+string.replace(wholeline," ","")+ShowHex(~chksum,8)+'\n') 
                    elif len(value) == 4:
                       chksum = 6+ int(value[0][:2],16)+int(value[0][2:],16)+int(value[1],16)+int(value[2],16)+int(value[3],16)
                       wholeline = (value[0]+' '+value[1]+' '+value[2]+' '+value[3]+("   "*1)) 
                       S019.write("S106"+string.replace(wholeline," ","")+ShowHex(~chksum,8)+'\n') 
                    elif len(value) == 5:
                       wholeline = (value[0]+' '+value[1]+' '+value[2]+' '+value[3]+' '+value[4])
                       chksum = 7+ int(value[0][:2],16)+int(value[0][2:],16)+int(value[1],16)+int(value[2],16)+int(value[3],16)+int(value[4],16)
                       S019.write("S107"+string.replace(wholeline," ","")+ShowHex(~chksum,8)+'\n') 
                    print wholeline+ErrorFlag+ExpandFlag+nextline
                    listing.write(wholeline+ErrorFlag+ExpandFlag+nextline+'\n')       
            else:
               wholeline = (" ..."+" "*17+ExpandFlag+nextline)
               print wholeline
               listing.write(wholeline+'\n')
###########################
        else:
           wholeline = (" "*21+ExpandFlag+nextline)
           print wholeline
           listing.write(wholeline+'\n')
        if ErrorFlag.startswith("-->"):
           ErrorCount = ErrorCount +1
    src.close()
    listing.close()
    S019.write("S9030000FC"+'\n')
    S019.close()
#    print("Errors:",ErrorCount)
    if ErrorCount ==0: 
       print('    There were no errors')
    elif ErrorCount == 1:
       print('    There was an error.')
    else:
       print('    There were '+str(ErrorCount)+' errors.')
#        print '***'



